#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <math.h>
#include <stdio.h>
#include <assert.h>
#include "Args.h"
#include "MaRatioData.h"
#include "TwoStateSegmentation.h"
#include "PriorParam.h"
#include "MCEM.h"
#include "utils.h"
#include <mpi.h>

using namespace std;

//put the file list in the text file(with each line contains a filename) into a vector
string extractflist(string fname, vector<string>& flist){
	ifstream fs(fname.c_str());
	string path="";
	assert(fs);
	string line;
	while (getline(fs,line) && line!=""){
		int i=line.find_last_of("/");
		if (i!=string::npos){
			path=line.substr(0,i+1);
			line.erase(0,i+1);
		}
		flist.push_back(line);
	}
	return path;
}

//split the file path/fname into short pieces, appended .split to each resulting file,
//split count starting from 1. n is the size of desired splitted files(with the exception
//of the last splitted file, if it is too short, will be combined with previous splitted file
//(currently, it is too short if length<min(n/2,2000). The arg skip is used when there is
//some header information contained in the file( in general do not need to set this if no header)
//return: number of splitted files in the pathsplit dir
int splitFile(string path,string pathsplit,string fname,int n, int skip=0){
	int counter=0;
	int nsplit=0;
	ifstream in((path+fname).c_str());
	assert(in);
	string line;
	//skip header
	for(int i=0;i<skip;i++) { getline(in,line);}
	while (in){
		nsplit++; counter=0;
		ofstream out((pathsplit+fname+".split"+int2string(nsplit)).c_str());
		while(counter<n && getline(in,line) ){
			counter++;
			out<<line<<endl;
		}
	}
	//if the final file is small, combine with the last file. But if this is the only file,
	//nothing to combine
	if (counter<2000 && counter<n/2 && nsplit>1){
		cerr<<"combining "<<nsplit<<" with "<<nsplit-1<<endl;
		ifstream from((pathsplit+fname+".split"+int2string(nsplit)).c_str());
		ofstream to((pathsplit+fname+".split"+int2string(nsplit-1)).c_str(),ios::app);
		to<<from.rdbuf();
		nsplit--;
		const char* removef=(pathsplit+fname+".split"+int2string(nsplit+1)).c_str();
		remove(removef);
	}
	cerr<<"splitting into "<<nsplit<<" subfiles"<<endl;
	return nsplit;

}


//concatenate files in dir path with names: fnames.suffix.1 through fnames.suffix.n into
//one single file named fnames.suffix, in the same dir
void mergeFile(string path,string fname,int n,string suffix=".postmean"){
	ofstream to((path+fname+suffix).c_str());
	for (int i=1;i<=n;i++){
		ifstream from((path+fname+suffix+int2string(i)).c_str());
		assert(from);
		to<<from.rdbuf();
	}
}


int main(int argc, char* argv[])
{
  //MPI: uncomment any comment that starts with MPI: if you have multiple segmentation files
  //and want to use MPI, currently MPI is not used for training step
  //MPI: MPI_Init(&argc,&argv);
  /*
   nUpdate: number of samples to be drawn during segmentation step to calculate centroid
   fsize: default size of splitted files
   emiter: default number of iterations for EM
  */
  int nUpdate=500;
  int fsize=500;
  int emiter=10;
  
  //initialize commandline arguments
  Args args(argc,argv);
  if (args.isset('n')) nUpdate=atoi(args.value('n').c_str());
  if (args.isset('s')) fsize=atoi(args.value('s').c_str());
  if (args.isset('e')) emiter=atoi(args.value('e').c_str());


  string pathsplit="";
  if (args.nargs()==4) pathsplit=args.arg(3);
  string priorfile=args.arg(0);
  string trainfilelist=args.arg(1);
  string filelist=args.arg(2);
  

  //read in intial hyperparameters
  PriorParam initprior;
  initprior=PriorParam(priorfile);
  
  
  //start training step
  vector<string> trainflist;
  string trainpathraw=extractflist(trainfilelist,trainflist);//after this step, trainflist does not contain path
  cerr<<"trainpath:"<<trainpathraw<<" file:"<<trainflist[0]<<endl;
  vector<string> trainsplitlist;
  //split files in trainsflist, resulting splitted file names stored in trainsplitlist
  for (vector<string>::iterator iter=trainflist.begin();iter!=trainflist.end();iter++){
    int nsplit=splitFile(trainpathraw,pathsplit,*iter,fsize);
    for(int i=1;i<=nsplit;i++){
	trainsplitlist.push_back(*iter+".split"+int2string(i));
    }
  }
  //start EM, return new hyperparameters after emiter iterations
  MCEM em(initprior,args,trainsplitlist,emiter);
  initprior=em.start();
  
 

  //start segmentation step
  vector<string> flist;
  string pathraw= extractflist(filelist,flist);//flist does not contain path anymore
  cerr<<"path:"<<pathraw<<" file:"<<flist[0]<<endl;
  string pathresult=pathraw;
  cout<<"number of files:"<<flist.size()<<endl;
  vector<string>::const_iterator fiter=flist.begin();
  
  //MPI:int myid, numprocs;
  //MPI:MPI_Comm_rank(MPI_COMM_WORLD,&myid);
  //MPI:MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
  int fileindex=-1;
  //first for below is for each orginal input file, 2nd for is loop over each splitted file
  for(;fiter!=flist.end();fiter++){
          fileindex++;
	  //MPI:if (fileindex%numprocs!=myid) continue;
	  string dataFileName=*fiter;
	  cout<<dataFileName<<endl;
	  int nsplit;
	  nsplit=splitFile(pathraw,pathsplit,dataFileName,fsize);
  	  for (int i=1;i<=nsplit;i++){
	  	//MPI:cerr<<"Node "<<myid<<" is processing subfile "<<i<<" for file "<<dataFileName<<":\n";
		cerr<<"processing subfile "<<i<<" for "<<dataFileName<<":\n";  
		MaRatioData mrdata;
		string fsplitname=pathsplit+dataFileName+".split"+int2string(i);
	  	mrdata=MaRatioData(fsplitname);
	  	TwoStateSegmentation segmenter(mrdata,args,initprior);
	  	segmenter.forwardRecursion();
	  	segmenter.computePosteriorKD();

	        /**********
	        ** output posterior probability of K and D(ending state)
	        **********/
	 	cerr<<"posterior probability of K and ending state:"<<endl;
	 	cerr<<segmenter.kmax<<endl;
	 	for (int k=0;k<=segmenter.kmax;k++){
		  cout<<"k="<<k<<" ";
		  for (int state=0;state<initprior.statesize;state++)
			  cout<<segmenter.posteriorKD[k][state]<<" ";
		  cout<<endl;
		}

	  ofstream fcentroid((pathsplit+dataFileName+".centroid"+int2string(i)).c_str(),ios::out);
          //the following commented output is not currently implemented in current version
	  //ofstream fupper((pathresult+dataFileName+".upper"+int2string(i)).c_str(),ios::out);
	  //ofstream flower((pathresult+dataFileName+".lower"+int2string(i)).c_str(),ios::out);
	  ofstream fcertainty((pathsplit+dataFileName+".certainty"+int2string(i)).c_str(),ios::out);
	  ofstream fpostmean((pathsplit+dataFileName+".postmean"+int2string(i)).c_str(),ios::out);
	  ofstream fpostvar((pathsplit+dataFileName+".postvar"+int2string(i)).c_str(),ios::out);
	  ofstream fstdres((pathsplit+dataFileName+".stdres"+int2string(i)).c_str(),ios::out);
	  ofstream fobs((pathsplit+dataFileName+".obs"+int2string(i)).c_str(),ios::out);
	  cerr<<"computing centroid and uncertainty"<<endl;
	  for(int i=1;i<=nUpdate;i++){
	      if (i%100==0) cerr<<i<<" ";
	  	  list<int>cutPts;
	  	  list<int>states;
		  //cutPts will contain last point of each seg, with states assignments in the list states
		  //also returns log probability
	  	  double probability=segmenter.backwardSampling(cutPts,states);
		  //store state assignments for this sample, used for computing centroid below
	  	  segmenter.updateVote(cutPts,states);	
		  segmenter.updatePostmeanvar(cutPts,states);
	  }
	  segmenter.computeCentroid();
	  segmenter.computeUncertainty();
	  segmenter.computeStdres();
	  segmenter.printCentroid(fcentroid);
	  //segmenter.printCentroidUpperUncertainty_5_col(fupper);
	  //segmenter.printCentroidLowerUncertainty_5_col(flower);
	  segmenter.printPostmean(fpostmean);
	  segmenter.printPostvar(fpostvar);
	  segmenter.printStdres(fstdres);
          segmenter.printCertainty(fcertainty);
          segmenter.printObs(fobs);
    }//for nsplit

    mergeFile(pathsplit,dataFileName,nsplit,".centroid");
    mergeFile(pathsplit,dataFileName,nsplit,".certainty");
    mergeFile(pathsplit,dataFileName,nsplit,".obs");
    mergeFile(pathsplit,dataFileName,nsplit,".postmean");
    mergeFile(pathsplit,dataFileName,nsplit,".postvar");
    mergeFile(pathsplit,dataFileName,nsplit,".stdres");
    //final output is .wig file
    ofstream fwig((pathresult+dataFileName+".wig").c_str(),ios::out);
    ifstream fobs((pathsplit+dataFileName+".obs").c_str());
    ifstream fcentroid((pathsplit+dataFileName+".centroid").c_str());
    ifstream fcertainty((pathsplit+dataFileName+".certainty").c_str());
    ifstream fpostmean((pathsplit+dataFileName+".postmean").c_str());
    ifstream fpostvar((pathsplit+dataFileName+".postvar").c_str());
    ifstream fstdres((pathsplit+dataFileName+".stdres").c_str());
    fwig<<"track type=wiggle_0 name=dec04-ngen1 description=dec04-ngen1"<<endl;
    fwig<<fobs.rdbuf();
    fwig<<"track type=wiggle_0 name=centroid description=dec04-ngen1"<<endl;
    fwig<<fcentroid.rdbuf();
    fwig<<"track type=wiggle_0 name=certainty description=dec04-ngen1"<<endl;
    fwig<<fcertainty.rdbuf();
    fwig<<"track type=wiggle_0 name=postmean description=dec04-ngen1"<<endl;
    fwig<<fpostmean.rdbuf();
    fwig<<"track type=wiggle_0 name=postvar description=dec04-ngen1"<<endl;
    fwig<<fpostvar.rdbuf();
    fwig<<"track type=wiggle_0 name=stdres description=dec04-ngen1"<<endl;
    fwig<<fstdres.rdbuf();
  }
  return 0;

}
