#ifndef UTILS_H
#define UTILS_H

#include <iostream>
#include <fstream>
#include <sstream>
#include <math.h>
#include <assert.h>
#include <vector>
#include <string>
#include <map>
#include <time.h>
#include <stdlib.h>
#include "Site.h"

#define SHIFT 4
using namespace std;

double computeCombinatorics(int n,int k,int min,int max);
void warning(const string& message="");
double lognchoosek(int n, int k);
string int2string(int i);
double lg(double nu,double sig);//compute the log of the coeffient before the density
								//of inv-chisq
pair<int,int> range(int i,int lower,int upper);
bool inRange(int j,int i,int lower,int upper);
void monitor(double d,const string& message="");
int multiSampling(double* prob,int size);
string ridComment(const string& s);
double loggammapdf(double x,double alpha,double beta);


template<class T>
T sum(int start,int end,const vector<T> & data){
	T res=T(0);
	int n=numobs(start,end,data);
	if (n==0) return T(MISSING-1);
	for (int i=start; i<=end; i++) if (data[i]>MISSING) res+=data[i];
	return res;
}

template<class T>
T sumsq(int start,int end,const vector<T>& data){
	T res=T(0);
	for (int i=start;i<=end;i++) if (data[i]>MISSING) res+=data[i]*data[i];
	return res;
}

template<class T>
T mean(int start,int end,const vector<T> & data){
	int n=numobs(start,end,data);
	if (n==0) return T(MISSING-1);
	return sum(start,end,data)/n;
}

template<class T>
T variance(int start, int end, const vector<T> & data){
	int n=numobs(start,end,data);
	if (n==1) return T(0);
	if (n==0) return T(MISSING-1);
	T datamean=mean(start,end,data);
	T datasq=T(0);
	for(int i=start;i<=end;i++){
		if (data[i]>MISSING)
		  datasq+=(data[i]-datamean)*(data[i]-datamean);
	}
	//assert(datasq>0);
	return datasq/(n-1);
}

template<class T>
int numobs(int start,int end, const vector<T>& data){
	int count=0;
	for(int i=start;i<=end;i++)
		if (data[i]>MISSING) count++;
	return count;
}



#endif
